from pasteob import *
from pasteob.auth.logic import AuthLogic, User
from mext.uri.encode import urlquote
from mext.uri.parse import unquote
from time import time
from hashlib import md5


__all__ = ['AuthTktLogic']


class AuthTktLogic(AuthLogic):
    secret = None
    cookie_name = 'auth_tkt'
    include_ip = True
    secure = False
    http_only = False
    persistent = False

    populate_req = True

    def __init__(self, req):
        assert self.secret, "Define auth_tkt secret in a subclass"
        # verified remote address
        self.remote_addr = req.remote_addr if self.include_ip else '0.0.0.0'
        # signals if there IS a cookie, but it didn't pass the check
        # if that's the case one should call .flush_cookies(resp)
        self.invalid_cookie = False
        self._parse_ticket(req) #@@ timeout
        cur_domain = req.host or req.server_name
        wild_domain = '.' + cur_domain
        self.domains = [None, cur_domain, wild_domain]


    def _parse_ticket(self, req):
        cookie = req.cookies.get(self.cookie_name, None)
        if not cookie:
            return
        ticket = AuthTicket.parse(self.secret, self.remote_addr, cookie)
        if not ticket:
            self.invalid_cookie = True
            return
        #self.ticket = ticket
        self.logged_in = True
        self._user = User(ticket.user_id, roles=ticket.tokens)
        self.user_data = ticket.user_data
        if self.populate_req:
            req.remote_user = ticket.user_id
            #req.environ['AUTH_TYPE'] = 'cookie'


    def do_logout(self, resp):
        assert self.logged_in
        self.flush_cookies(resp)

    def flush_cookies(self, resp):
        for domain in self.domains:
            resp.delete_cookie(self.cookie_name, domain=domain)

    def do_login(self, resp, user, user_data=''):
        assert not self.logged_in
        if user.email:
            raise ValueError("AuthTktLogic.do_login is not capable of remembering user emails")
        elif user.nickname != user.user_id:
            raise ValueError("AuthTktLogic.do_login is incapable of remembering users with different user_id and nickname")
        #@@ refresh
        self.set_user_cookie(resp, user.user_id, sorted(user.roles), user_data)


    # Be careful not to do an HTTP redirect after login; use meta
    # refresh or Javascript -- some browsers have bugs where cookies
    # aren't saved when set on a redirect.

    def set_user_cookie(self, resp, user_id, tokens='', user_data=''):
        if not isinstance(tokens, basestring):
            tokens = ','.join(tokens)
        ticket = AuthTicket(self.secret, self.remote_addr, time(), user_id, tokens, user_data)
        value = ticket.cookie_value
        for domain in self.domains:
            resp.set_cookie(
                self.cookie_name, value,
                domain=domain,
                persistent=self.persistent,
                secure=self.secure,
                httponly=self.http_only,
            )










## class BadTicket(Exception):
##     pass
##

class AuthTicket(object):
    def __init__(self, secret, ip, timestamp, user_id, tokens=(), user_data=''):
        self.ip = ip
        self.timestamp = timestamp
        self.user_id = user_id
        self.tokens = tokens
        self.user_data = user_data or ''
        self.digest = calculate_digest(ip, timestamp, secret, user_id, ','.join(tokens), user_data)

    @classmethod
    def parse(cls, secret, ip, ticketstr):
        ticketstr = ticketstr.strip('"')
        digest = ticketstr[:32]
        try:
            timestamp = int(ticketstr[32:40], 16)
        except ValueError:
            return
            #raise BadTicket
        parts = ticketstr[40:].split('!')
        user_id = unquote(parts.pop(0))
        tokens, user_data = (), ''
        if parts:
            if len(parts) == 2:
                tokens = map(unquote, parts.pop(0).split(','))
            user_data = unquote(parts.pop())
        ticket = cls(secret, ip, timestamp, user_id, tokens, user_data)
        if ticket.digest != digest:
            return
            #raise BadTicket('Digest signature is not correct', ticket.digest, digest)
        return ticket

    @property
    def cookie_value(self):
        r = ['%s%08x%s' % (
                self.digest,
                int(self.timestamp),
                urlquote(self.user_id, '!')
            ),
            ','.join(urlquote(t, '!,') for t in self.tokens),
            urlquote(self.user_data, '!')
        ]
        return '!'.join(filter(None, r))

def calculate_digest(ip, timestamp, secret, user_id, tokens, user_data):
    secret = maybe_encode(secret)
    user_id = maybe_encode(user_id)
    tokens = maybe_encode(tokens)
    user_data = maybe_encode(user_data)
    digest0 = md5(
        encode_ip_timestamp(ip, timestamp) + secret + user_id + '\0'
        + tokens + '\0' + user_data).hexdigest()
    digest = md5(digest0 + secret).hexdigest()
    return digest

def encode_ip_timestamp(ip, timestamp):
    ip = map(int, ip.split('.'))
    t = int(timestamp)
    ts = [
        (t & 0xff000000) >> 24,
        (t & 0x00ff0000) >> 16,
        (t & 0x0000ff00) >> 8,
        (t & 0x000000ff)
    ]
    return ''.join(map(chr, ip+ts))

def maybe_encode(s, encoding='utf8'):
    if isinstance(s, unicode):
        s = s.encode(encoding)
    assert isinstance(s, str)
    return s

