from webob.response import iter_close
from pasteob.base import *
from pasteob.errors import *
from warnings import warn
from datetime import timedelta

__all__ = [
    'fix_HEAD', 'override_https', 'override_errors', 'chrome_frame_mw',
    'Cascade', 'SiteRoot'
]


@wapp_mw
def fix_HEAD(req, app):
    """
        HEAD requests must return exactly the same headers as a GET.
        Most apps / dispatchers / middleware do not honour this, so
        we have to cheat them.
    """
    if req.method == 'HEAD':
        req.method = 'GET'
        app = app << req
        iter_close(app.app_iter)
        app.app_iter = ()
        req.method = 'HEAD'
    return app


@wapp_mw
def override_https(req, normal_app, secure_app):
    """
        Return a WSGI app that dispatches HTTP requests to normal_app and HTTPS to secure_app.
        Use override_https('redirect', secure_app) to force HTTPS connections.
    """
    if normal_app == 'redirect' and req.scheme == 'http':
        return RedirectResponse('permanent', 'https' + req.url[4:])
    return secure_app if req.scheme == 'https' else normal_app


@wapp_mw
def override_errors(req, app, overrides={}):
   r = app << req
   return overrides.get(r.status_int, r)

@wapp_mw
def chrome_frame_mw(req, app):
    r = app << req
    if 'html' in (r.content_type or ''):
        r.headers['X-UA-Compatible'] = 'chrome=1'
    return r



def Cascade(apps, catch=(404,)):
    if len(apps) == 0:
        raise ValueError("Need at least one app for the cascade")
    elif len(apps) == 1:
        return apps[0]
    for item in catch:
        assert isinstance(item, int)
    catch = frozenset(catch)

    @wapp
    def cascade_mw(req):
        for app in apps:
            resp = app << req.copy()
            if resp.status_int not in catch:
                break
        return resp
    return cascade_mw



class SiteRoot(object):
    overrides = {
        '/robots.txt': 'robots',
        '/favicon.ico': 'favicon',
        '/crossdomain.xml': 'crossdomain',
    }

    default_expires = timedelta(days=3)

    def __init__(self, app):
        self.app = app
        self.favicon = ErrorResponse(404, cache_expires=self.default_expires)
        self.robots = Response(
            ctype='text/plain',
            cache_expires=self.default_expires
        )
        self.crossdomain = Response(
            _crossdomain_deny,
            ctype='application/xml',
            cache_expires=self.default_expires
        )
        self.overrides = self.overrides.copy()

    @wapp
    def __call__(self, req):
        if req.script_name:
            warn("SiteRoot should be mounted at site root")
        if req.path_info in self.overrides:
            app = self.overrides[req.path_info]
            if isinstance(app, str):
                app = getattr(self, app)
            return app
        return self.app

_crossdomain_deny = '''<?xml version="1.0"?>
<!DOCTYPE cross-domain-policy SYSTEM "http://www.macromedia.com/xml/dtds/cross-domain-policy.dtd">
<cross-domain-policy><site-control permitted-cross-domain-policies="none"/></cross-domain-policy>'''

