import sys
from cStringIO import StringIO
from urllib import unquote #@@
from pasteob.errors import ErrorResponse
from webob.headers import _trans_name as _header_key
try:
    from mext.net.rfile import SockRFile
except ImportError:
    from weakref import WeakKeyDictionary
    _rfile = WeakKeyDictionary()
    def SockRFile(sock):
        return _rfile.setdefault(sock, sock.makefile('rb'))




__all__ = ['parse_request']

class CloseConnection(BaseException):
    pass

class HeaderError(ValueError):
    pass

CRLF = '\r\n'



def parse_request(conn):
    rfile = SockRFile(conn)
    line0 = rfile.readline()
    if not line0 or not line0.endswith(CRLF):
        raise CloseConnection
    line0 = line0[:-2]
    try:
        (method, path, ver) = line0.split()
    except:
        return ErrorResponse(400, "Bad request syntax: %s" % line0)
    try:
        if not ver.startswith('HTTP/'):
            raise ValueError
        ver_no = tuple(map(int, ver.split('/', 1)[1].split(".")))
        if len(ver_no) != 2:
            raise ValueError
    except (ValueError, IndexError):
        return ErrorResponse(400, "Bad request version: %s" % ver)
    if ver_no[0] != 1:
        return ErrorResponse(505, "Unsupported HTTP version: %s" % ver)

    if '?' in path:
        path, query = path.split('?', 1)
    else:
        query = ''

    env = {
        'REQUEST_METHOD': method,
        'SCRIPT_NAME': '',
        'PATH_INFO': unquote(path),
        'QUERY_STRING': query, # wsgiref and cherrypy seem not to unquote the query
        #'QUERY_STRING': unquote(query),
        #'SERVER_PROTOCOL': 'HTTP/1.1',
        'SERVER_PROTOCOL': ver,
        'REMOTE_ADDR': conn.getpeername()[0],
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': 'http',
        'wsgi.errors': sys.stderr,
        'wsgi.multithread': True,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
    }

    try:
        read_headers(rfile, env)
    except HeaderError, exc:
        return ErrorResponse(400, exc[0])

    if method in ('POST', 'PUT'):
        clen = env.get('CONTENT_LENGTH')
        if clen is None:
            return ErrorResponse(400, "%s requests should supply content-length" % method)
        if env.get('HTTP_EXPECT', '').lower() == '100-continue':
            rfile = ContinueHook(rfile, conn)
        rfile = LimitedLengthFile(rfile, int(clen))
    else:
        rfile = StringIO('')

    env['wsgi.input'] = rfile
    return env

## comma_separated_headers = map(_header_key,
##     ['Accept', 'Accept-Charset',  'Accept-Encoding',
##     'Accept-Language', 'Accept-Ranges', 'Allow', 'Cache-Control',
##     'Connection', 'Content-Encoding', 'Content-Language', 'Expect',
##     'If-Match', 'If-None-Match', 'Pragma', 'Proxy-Authenticate', 'TE',
##     'Trailer', 'Transfer-Encoding', 'Upgrade', 'Vary', 'Via', 'Warning',
##     'WWW-Authenticate'
## ])
##

def read_headers(rfile, env): # adapted from cherrypy
    rfile = LimitedLengthFile(rfile, 1<<12) # max headers size: 4K
    for line in rfile:
        if line == CRLF: # end of headers
            return
        elif not line.endswith(CRLF): # bad line terminator
            raise HeaderError("HTTP requires CRLF terminators")
        try:
            k, v = line.split(":", 1)
        except ValueError:
            raise HeaderError("Bad header: %s" % line)
        k = _header_key(k)
        v = v.strip()
        if k in env:# and k in comma_separated_headers:
            env[k] += ', ' + v
        else:
            env[k] = v
    else:
        # no more data
        raise HeaderError("Unexpected end of headers")


class ContinueHook(object):
    def __init__(self, rfile, conn):
        self._rfile = rfile
        self._conn = conn

##     def __getattr__(self, name):
##         return getattr(self._rfile, name)
##

    def _send_continue(self):
        self._conn.sendall("HTTP/1.1 100 Continue\r\n\r\n")
        rfile = self._rfile
        for attr in ('read', 'readline', 'readlines', 'next'):
            setattr(self, attr, getattr(rfile, attr))

    def read(self, size=-1):
        self._send_continue()
        return self._rfile.read(size)

    def readline(self, size=-1):
        self._send_continue()
        return self._rfile.readline(size)

    def readlines(self, sizehint=0):
        self._send_continue()
        return self._rfile.readlines(sizehint)

    def __iter__(self):
        return self

    def next(self):
        self._send_continue()
        return self._rfile.next()


class LimitedLengthFile(object):
    def __init__(self, file, maxlen):
        self.file = file
        self.maxlen = maxlen
        self.remaining = maxlen

    def __repr__(self):
        base_repr = repr(self.file)
        return base_repr[:-1] + ' maxlen=%s>' % self.maxlen

    def read(self, sz=-1):
        if sz is None or sz < 0:
            sz = self.remaining
        else:
            sz = min(sz, self.remaining)
        if not sz:
            return ''
        r = self.file.read(sz)
        self.remaining -= len(r)
        return r

    def readline(self, hint=None):
        if hint:
            hint = min(hint, self.remaining)
        else:
            hint = self.remaining
        r = self.file.readline(hint)
        self.remaining -= len(r)
        return r

    def __iter__(self):
        return self

    def next(self):
        r = self.readline()
        if not r:
            raise StopIteration
        return r
